@@ -4,18 +4,33 @@
 
 #include <time.h>
 #include <string.h>
-#include <dlfcn.h>
 
-#ifndef RTLD_DEFAULT
-#define RTLD_DEFAULT ((void*)0)
+/* We will want to check for symbols in the OpenCL library.
+ * On Windows, we must get the module handle for it, on Unix-like
+ * systems we can just use RTLD_DEFAULT
+ */
+#ifdef _MSC_VER
+# include <windows.h>
+# define dlsym GetProcAddress
+# define DL_MODULE GetModuleHandle("OpenCL")
+#else
+# include <dlfcn.h>
+# define DL_MODULE ((void*)0) /* This would be RTLD_DEFAULT */
 #endif
 
+/* ISO C forbids assignments between function pointers and void pointers,
+ * but POSIX allows it. To compile without warnings even in -pedantic mode,
+ * we use this horrible trick to get a function address from
+ * clGetExtensionFunctionAddress
+ */
+#define PTR_FUNC_PTR *(void**)&
+
 /* Load STDC format macros (PRI*), or define them
  * for those crappy, non-standard compilers
  */
 #include "fmtmacros.h"
 
-// Support for the horrible MS C compiler
+// More support for the horrible MS C compiler
 #ifdef _MSC_VER
 #include "ms_support.h"
 #endif
@@ -26,7 +41,10 @@
 #include "strbuf.h"
 
 #define ARRAY_SIZE(ar) (sizeof(ar)/sizeof(*ar))
-#define UNUSED __attribute__((unused))
+
+#ifndef UNUSED
+#define UNUSED(x) x __attribute__((unused))
+#endif
 
 struct platform_data {
 	char *pname; /* CL_PLATFORM_NAME */
@@ -102,6 +120,8 @@ static const cl_device_type devtype[] = { 0,
 	CL_DEVICE_TYPE_ACCELERATOR, CL_DEVICE_TYPE_CUSTOM, CL_DEVICE_TYPE_ALL };
 
 const size_t devtype_count = ARRAY_SIZE(devtype);
+/* number of actual device types, without ALL */
+const size_t actual_devtype_count = ARRAY_SIZE(devtype) - 1;
 
 static const char* device_type_str[] = { unk, "Default", "CPU", "GPU", "Accelerator", "Custom", "All" };
 static const char* device_type_raw_str[] = { unk,
@@ -155,6 +175,16 @@ static const char* affinity_domain_raw_ext_str[] = {
 
 const size_t affinity_domain_count = ARRAY_SIZE(affinity_domain_str);
 
+static const char *terminate_capability_str[] = {
+	"Context"
+};
+
+static const char *terminate_capability_raw_str[] = {
+	"CL_DEVICE_TERMINATE_CAPABILITY_CONTEXT_KHR"
+};
+
+const size_t terminate_capability_count = ARRAY_SIZE(terminate_capability_str);
+
 static const char* fp_conf_str[] = {
 	"Denormals", "Infinity and NANs", "Round to nearest", "Round to zero",
 	"Round to infinity", "IEEE754-2008 fused multiply-add",
@@ -191,11 +221,15 @@ static const char* svm_cap_raw_str[] = {
 
 const size_t svm_cap_count = ARRAY_SIZE(svm_cap_str);
 
+/* SI suffixes for memory sizes. Note that in OpenCL most of them are
+ * passed via a cl_ulong, which at most can mode 16 EiB, but hey,
+ * let's be forward-thinking ;-)
+ */
 static const char* memsfx[] = {
-	"B", "KiB", "MiB", "GiB", "TiB"
+	"B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB"
 };
 
-const size_t memsfx_count = ARRAY_SIZE(memsfx);
+const size_t memsfx_end = ARRAY_SIZE(memsfx) + 1;
 
 static const char* lmem_type_str[] = { none, "Local", "Global" };
 static const char* lmem_type_raw_str[] = { none_raw, "CL_LOCAL", "CL_GLOBAL" };
@@ -306,18 +340,9 @@ void show_strbuf(const char *pname, int skip)
 }
 
 int
-platform_info_str(cl_platform_id pid, cl_platform_info param, const char* pname, const struct platform_info_checks * chk UNUSED)
+platform_info_str(cl_platform_id pid, cl_platform_info param, const char* pname, const struct platform_info_checks* UNUSED(chk))
 {
-	error = clGetPlatformInfo(pid, param, 0, NULL, &nusz);
-	if (nusz > bufsz) {
-		REALLOC(strbuf, nusz, current_param);
-		bufsz = nusz;
-	}
-	had_error = REPORT_ERROR2("get %s size");
-	if (!had_error) {
-		error = clGetPlatformInfo(pid, param, bufsz, strbuf, NULL);
-		had_error = REPORT_ERROR2("get %s");
-	}
+	GET_STRING2(clGetPlatformInfo, pid, param);
 	/* when only listing, do not print anything we're just gathering
 	 * information
 	 */
@@ -327,7 +352,7 @@ platform_info_str(cl_platform_id pid, cl_platform_info param, const char* pname,
 }
 
 int
-platform_info_ulong(cl_platform_id pid, cl_platform_info param, const char* pname, const struct platform_info_checks * chk UNUSED)
+platform_info_ulong(cl_platform_id pid, cl_platform_info param, const char* pname, const struct platform_info_checks* UNUSED(chk))
 {
 	cl_ulong val = 0;
 
@@ -501,11 +526,18 @@ getWGsizes(cl_platform_id pid, cl_device_id dev)
 
 	/* for a program build failure, dump the log to stderr before bailing */
 	if (error == CL_BUILD_PROGRAM_FAILURE) {
+		/* Do not clobber strbuf, shadow it */
+		char *strbuf = NULL;
+		size_t bufsz = 0, nusz = 0;
 		GET_STRING(clGetProgramBuildInfo, CL_PROGRAM_BUILD_LOG, "CL_PROGRAM_BUILD_LOG", prg, dev);
 		if (error == CL_SUCCESS) {
+			fflush(stdout);
+			fflush(stderr);
 			fputs("=== CL_PROGRAM_BUILD_LOG ===\n", stderr);
 			fputs(strbuf, stderr);
+			fflush(stderr);
 		}
+		free(strbuf);
 	}
 	if (had_error)
 		goto out;
@@ -542,20 +574,27 @@ struct device_info_checks {
 	cl_device_mem_cache_type cachetype;
 	cl_device_local_mem_type lmemtype;
 	cl_bool image_support;
+	cl_bool compiler_available;
 	char has_half[12];
 	char has_double[24];
 	char has_nv[29];
 	char has_amd[30];
-	char has_svm_ext[11];
+	char has_amd_svm[11];
+	char has_arm_svm[29];
 	char has_fission[22];
 	char has_atomic_counters[26];
 	char has_image2d_buffer[27];
 	char has_intel_local_thread[30];
 	char has_intel_AME[36];
+	char has_intel_AVC_ME[43];
+	char has_intel_planar_yuv[20];
+	char has_intel_required_subgroup_size[32];
 	char has_altera_dev_temp[29];
 	char has_spir[12];
 	char has_qcom_ext_host_ptr[21];
 	char has_simultaneous_sharing[30];
+	char has_subgroup_named_barrier[30];
+	char has_terminate_context[25];
 	cl_uint dev_version;
 };
 
@@ -568,16 +607,22 @@ DEFINE_EXT_CHECK(half)
 DEFINE_EXT_CHECK(double)
 DEFINE_EXT_CHECK(nv)
 DEFINE_EXT_CHECK(amd)
-DEFINE_EXT_CHECK(svm_ext)
+DEFINE_EXT_CHECK(amd_svm)
+DEFINE_EXT_CHECK(arm_svm)
 DEFINE_EXT_CHECK(fission)
 DEFINE_EXT_CHECK(atomic_counters)
 DEFINE_EXT_CHECK(image2d_buffer)
 DEFINE_EXT_CHECK(intel_local_thread)
 DEFINE_EXT_CHECK(intel_AME)
+DEFINE_EXT_CHECK(intel_AVC_ME)
+DEFINE_EXT_CHECK(intel_planar_yuv)
+DEFINE_EXT_CHECK(intel_required_subgroup_size)
 DEFINE_EXT_CHECK(altera_dev_temp)
 DEFINE_EXT_CHECK(spir)
 DEFINE_EXT_CHECK(qcom_ext_host_ptr)
 DEFINE_EXT_CHECK(simultaneous_sharing)
+DEFINE_EXT_CHECK(subgroup_named_barrier)
+DEFINE_EXT_CHECK(terminate_context)
 
 /* In the version checks we negate the opposite conditions
  * instead of double-negating the actual condition
@@ -620,7 +665,7 @@ int dev_is_gpu_amd(const struct device_info_checks *chk)
 
 int dev_has_svm(const struct device_info_checks *chk)
 {
-	return dev_is_20(chk) || dev_has_svm_ext(chk);
+	return dev_is_20(chk) || dev_has_amd_svm(chk);
 }
 
 int dev_has_partition(const struct device_info_checks *chk)
@@ -653,6 +698,11 @@ int dev_has_images_20(const struct device_info_checks *chk)
 	return dev_has_images(chk) && dev_is_20(chk);
 }
 
+int dev_has_compiler(const struct device_info_checks *chk)
+{
+	return chk->compiler_available;
+}
+
 
 void identify_device_extensions(const char *extensions, struct device_info_checks *chk)
 {
@@ -677,7 +727,8 @@ void identify_device_extensions(const char *extensions, struct device_info_check
 		CHECK_EXT(double, cl_APPLE_fp64_basic_ops);
 	CHECK_EXT(nv, cl_nv_device_attribute_query);
 	CHECK_EXT(amd, cl_amd_device_attribute_query);
-	CHECK_EXT(svm_ext, cl_amd_svm);
+	CHECK_EXT(amd_svm, cl_amd_svm);
+	CHECK_EXT(arm_svm, cl_arm_shared_virtual_memory);
 	CHECK_EXT(fission, cl_ext_device_fission);
 	CHECK_EXT(atomic_counters, cl_ext_atomic_counters_64);
 	if (dev_has_atomic_counters(chk))
@@ -685,13 +736,17 @@ void identify_device_extensions(const char *extensions, struct device_info_check
 	CHECK_EXT(image2d_buffer, cl_khr_image2d_from_buffer);
 	CHECK_EXT(intel_local_thread, cl_intel_exec_by_local_thread);
 	CHECK_EXT(intel_AME, cl_intel_advanced_motion_estimation);
+	CHECK_EXT(intel_AVC_ME, cl_intel_device_side_avc_motion_estimation);
+	CHECK_EXT(intel_planar_yuv, cl_intel_planar_yuv);
+	CHECK_EXT(intel_required_subgroup_size, cl_intel_required_subgroup_size);
 	CHECK_EXT(altera_dev_temp, cl_altera_device_temperature);
 	CHECK_EXT(qcom_ext_host_ptr, cl_qcom_ext_host_ptr);
 	CHECK_EXT(simultaneous_sharing, cl_intel_simultaneous_sharing);
+	CHECK_EXT(subgroup_named_barrier, cl_khr_subgroup_named_barrier);
+	CHECK_EXT(terminate_context, cl_khr_terminate_context);
 }
 
 
-
 /*
  * Device info print functions
  */
@@ -735,7 +790,7 @@ void identify_device_extensions(const char *extensions, struct device_info_check
 
 #define DEFINE_DEVINFO_SHOW(how, type, fmt) \
 int device_info_##how(cl_device_id dev, cl_device_info param, const char *pname, \
-	const struct device_info_checks *chk UNUSED) \
+	const struct device_info_checks* UNUSED(chk)) \
 { \
 	type val = 0; \
 	SHOW_VAL(fmt); \
@@ -744,19 +799,10 @@ int device_info_##how(cl_device_id dev, cl_device_info param, const char *pname,
 
 /* Get string-type info without showing it */
 int device_info_str_get(cl_device_id dev, cl_device_info param, const char *pname,
-	const struct device_info_checks *chk UNUSED)
+	const struct device_info_checks* UNUSED(chk))
 {
 	current_param = pname;
-	error = clGetDeviceInfo(dev, param, 0, NULL, &nusz);
-	if (nusz > bufsz) {
-		REALLOC(strbuf, nusz, current_param);
-		bufsz = nusz;
-	}
-	had_error = REPORT_ERROR2("get %s size");
-	if (!had_error) {
-		error = clGetDeviceInfo(dev, param, bufsz, strbuf, NULL);
-		had_error = REPORT_ERROR2("get %s");
-	}
+	GET_STRING2(clGetDeviceInfo, dev, param);
 	return had_error;
 }
 
@@ -774,7 +820,7 @@ DEFINE_DEVINFO_SHOW(long, cl_ulong, "%" PRIu64)
 DEFINE_DEVINFO_SHOW(sz, size_t, "%" PRIuS)
 
 int device_info_bool(cl_device_id dev, cl_device_info param, const char *pname,
-	const struct device_info_checks *chk UNUSED)
+	const struct device_info_checks* UNUSED(chk))
 {
 	cl_bool val = 0;
 	const char * const * str = (output_mode == CLINFO_HUMAN ?
@@ -793,7 +839,7 @@ int device_info_bool(cl_device_id dev, cl_device_info param, const char *pname,
 }
 
 int device_info_bits(cl_device_id dev, cl_device_info param, const char *pname,
-	const struct device_info_checks *chk UNUSED)
+	const struct device_info_checks* UNUSED(chk))
 {
 	cl_uint val;
 	GET_VAL;
@@ -806,9 +852,9 @@ int device_info_bits(cl_device_id dev, cl_device_info param, const char *pname,
 
 size_t strbuf_mem(cl_ulong val, size_t szval)
 {
-	double dbl = val;
+	double dbl = (double)val;
 	size_t sfx = 0;
-	while (dbl > 1024 && sfx < memsfx_count) {
+	while (dbl > 1024 && sfx < memsfx_end) {
 		dbl /= 1024;
 		++sfx;
 	}
@@ -817,7 +863,7 @@ size_t strbuf_mem(cl_ulong val, size_t szval)
 }
 
 int device_info_mem(cl_device_id dev, cl_device_info param, const char *pname,
-	const struct device_info_checks *chk UNUSED)
+	const struct device_info_checks* UNUSED(chk))
 {
 	cl_ulong val = 0;
 	size_t szval = 0;
@@ -832,7 +878,7 @@ int device_info_mem(cl_device_id dev, cl_device_info param, const char *pname,
 }
 
 int device_info_mem_int(cl_device_id dev, cl_device_info param, const char *pname,
-	const struct device_info_checks *chk UNUSED)
+	const struct device_info_checks* UNUSED(chk))
 {
 	cl_uint val = 0;
 	size_t szval = 0;
@@ -846,8 +892,23 @@ int device_info_mem_int(cl_device_id dev, cl_device_info param, const char *pnam
 	return had_error;
 }
 
+int device_info_mem_sz(cl_device_id dev, cl_device_info param, const char *pname,
+	const struct device_info_checks* UNUSED(chk))
+{
+	size_t val = 0;
+	size_t szval = 0;
+	GET_VAL;
+	if (!had_error) {
+		szval += sprintf(strbuf, "%" PRIuS, val);
+		if (output_mode == CLINFO_HUMAN && val > 1024)
+			strbuf_mem(val, szval);
+	}
+	show_strbuf(pname, 0);
+	return had_error;
+}
+
 int device_info_free_mem_amd(cl_device_id dev, cl_device_info param, const char *pname,
-	const struct device_info_checks *chk UNUSED)
+	const struct device_info_checks* UNUSED(chk))
 {
 	size_t *val = NULL;
 	size_t szval = 0, numval = 0;
@@ -871,7 +932,7 @@ int device_info_free_mem_amd(cl_device_id dev, cl_device_info param, const char
 }
 
 int device_info_time_offset(cl_device_id dev, cl_device_info param, const char *pname,
-	const struct device_info_checks *chk UNUSED)
+	const struct device_info_checks* UNUSED(chk))
 {
 	cl_ulong val = 0;
 	GET_VAL;
@@ -889,7 +950,7 @@ int device_info_time_offset(cl_device_id dev, cl_device_info param, const char *
 }
 
 int device_info_szptr(cl_device_id dev, cl_device_info param, const char *pname,
-	const struct device_info_checks *chk UNUSED)
+	const struct device_info_checks* UNUSED(chk))
 {
 	size_t *val = NULL;
 	size_t szval = 0, numval = 0;
@@ -912,8 +973,8 @@ int device_info_szptr(cl_device_id dev, cl_device_info param, const char *pname,
 	return had_error;
 }
 
-int device_info_wg(cl_device_id dev, cl_device_info param UNUSED, const char *pname,
-	const struct device_info_checks *chk UNUSED)
+int device_info_wg(cl_device_id dev, cl_device_info UNUSED(param), const char *pname,
+	const struct device_info_checks* UNUSED(chk))
 {
 	cl_platform_id val = NULL;
 	{
@@ -933,7 +994,7 @@ int device_info_wg(cl_device_id dev, cl_device_info param UNUSED, const char *pn
 }
 
 int device_info_img_sz_2d(cl_device_id dev, cl_device_info param, const char *pname,
-	const struct device_info_checks *chk UNUSED)
+	const struct device_info_checks* UNUSED(chk))
 {
 	size_t width = 0, height = 0, val = 0;
 	GET_VAL; /* HEIGHT */
@@ -951,8 +1012,28 @@ int device_info_img_sz_2d(cl_device_id dev, cl_device_info param, const char *pn
 	return had_error;
 }
 
+int device_info_img_sz_intel_planar_yuv(cl_device_id dev, cl_device_info param, const char *pname,
+	const struct device_info_checks* UNUSED(chk))
+{
+	size_t width = 0, height = 0, val = 0;
+	GET_VAL; /* HEIGHT */
+	if (!had_error) {
+		height = val;
+		param = CL_DEVICE_PLANAR_YUV_MAX_WIDTH_INTEL;
+		current_param = "CL_DEVICE_PLANAR_YUV_MAX_WIDTH_INTEL";
+		GET_VAL;
+		if (!had_error) {
+			width = val;
+			sprintf(strbuf, "%" PRIuS "x%" PRIuS, width, height);
+		}
+	}
+	show_strbuf(pname, 0);
+	return had_error;
+}
+
+
 int device_info_img_sz_3d(cl_device_id dev, cl_device_info param, const char *pname,
-	const struct device_info_checks *chk UNUSED)
+	const struct device_info_checks* UNUSED(chk))
 {
 	size_t width = 0, height = 0, depth = 0, val = 0;
 	GET_VAL; /* HEIGHT */
@@ -979,16 +1060,14 @@ int device_info_img_sz_3d(cl_device_id dev, cl_device_info param, const char *pn
 
 
 int device_info_devtype(cl_device_id dev, cl_device_info param, const char *pname,
-	const struct device_info_checks *chk UNUSED)
+	const struct device_info_checks* UNUSED(chk))
 {
 	cl_device_type val = 0;
 	GET_VAL;
 	if (!had_error) {
 		/* iterate over device type strings, appending their textual form
-		 * to strbuf.
-		 * TODO: check for extra bits/no bits
-		 */
-		cl_uint i = devtype_count - 1; /* skip CL_DEVICE_TYPE_ALL */
+		 * to strbuf */
+		cl_uint i = actual_devtype_count;
 		const char * const *devstr = (output_mode == CLINFO_HUMAN ?
 			device_type_str : device_type_raw_str);
 		size_t szval = 0;
@@ -1003,6 +1082,15 @@ int device_info_devtype(cl_device_id dev, cl_device_info param, const char *pnam
 				szval += bufcpy(szval, devstr[i]);
 			}
 		}
+		/* check for extra bits */
+		if (szval < bufsz) {
+			cl_device_type known_mask = ((cl_device_type)(1) << actual_devtype_count) - 1;
+			cl_device_type extra = val & ~known_mask;
+			if (extra) {
+				add_separator(&szval);
+				szval += snprintf(strbuf + szval, bufsz - szval - 1, "0x%" PRIX64, extra);
+			}
+		}
 	}
 	show_strbuf(pname, 0);
 	/* we abuse global strbuf to pass the device type over to the caller */
@@ -1012,7 +1100,7 @@ int device_info_devtype(cl_device_id dev, cl_device_info param, const char *pnam
 }
 
 int device_info_cachetype(cl_device_id dev, cl_device_info param, const char *pname,
-	const struct device_info_checks *chk UNUSED)
+	const struct device_info_checks* UNUSED(chk))
 {
 	cl_device_mem_cache_type val = 0;
 	GET_VAL;
@@ -1029,7 +1117,7 @@ int device_info_cachetype(cl_device_id dev, cl_device_info param, const char *pn
 }
 
 int device_info_lmemtype(cl_device_id dev, cl_device_info param, const char *pname,
-	const struct device_info_checks *chk UNUSED)
+	const struct device_info_checks* UNUSED(chk))
 {
 	cl_device_local_mem_type val = 0;
 	GET_VAL;
@@ -1070,7 +1158,7 @@ void devtopo_str(const cl_device_topology_amd *devtopo)
 }
 
 int device_info_devtopo_amd(cl_device_id dev, cl_device_info param, const char *pname,
-	const struct device_info_checks *chk UNUSED)
+	const struct device_info_checks* UNUSED(chk))
 {
 	cl_device_topology_amd val;
 	GET_VAL;
@@ -1084,7 +1172,7 @@ int device_info_devtopo_amd(cl_device_id dev, cl_device_info param, const char *
 
 /* we assemble a cl_device_topology_amd struct from the NVIDIA info */
 int device_info_devtopo_nv(cl_device_id dev, cl_device_info param, const char *pname,
-	const struct device_info_checks *chk UNUSED)
+	const struct device_info_checks* UNUSED(chk))
 {
 	cl_device_topology_amd devtopo;
 	cl_uint val = 0;
@@ -1102,7 +1190,7 @@ int device_info_devtopo_nv(cl_device_id dev, cl_device_info param, const char *p
 		GET_VAL;
 
 		if (!had_error) {
-			devtopo.pcie.device = val >> 3;
+			devtopo.pcie.device = (val >> 3) & 0xff;
 			devtopo.pcie.function = val & 7;
 			devtopo_str(&devtopo);
 		}
@@ -1114,7 +1202,7 @@ int device_info_devtopo_nv(cl_device_id dev, cl_device_info param, const char *p
 
 /* NVIDIA Compute Capability */
 int device_info_cc_nv(cl_device_id dev, cl_device_info param, const char *pname,
-	const struct device_info_checks *chk UNUSED)
+	const struct device_info_checks* UNUSED(chk))
 {
 	cl_uint major = 0, val = 0;
 	GET_VAL; /* MAJOR */
@@ -1133,7 +1221,7 @@ int device_info_cc_nv(cl_device_id dev, cl_device_info param, const char *pname,
 
 /* AMD GFXIP */
 int device_info_gfxip_amd(cl_device_id dev, cl_device_info param, const char *pname,
-	const struct device_info_checks *chk UNUSED)
+	const struct device_info_checks* UNUSED(chk))
 {
 	cl_uint major = 0, val = 0;
 	GET_VAL; /* MAJOR */
@@ -1152,7 +1240,7 @@ int device_info_gfxip_amd(cl_device_id dev, cl_device_info param, const char *pn
 
 
 /* Device Partition, CLINFO_HUMAN header */
-int device_info_partition_header(cl_device_id dev UNUSED, cl_device_info param UNUSED,
+int device_info_partition_header(cl_device_id UNUSED(dev), cl_device_info UNUSED(param),
 	const char *pname, const struct device_info_checks *chk)
 {
 	int is_12 = dev_is_12(chk);
@@ -1171,7 +1259,7 @@ int device_info_partition_header(cl_device_id dev UNUSED, cl_device_info param U
 
 /* Device partition properties */
 int device_info_partition_types(cl_device_id dev, cl_device_info param, const char *pname,
-	const struct device_info_checks *chk UNUSED)
+	const struct device_info_checks* UNUSED(chk))
 {
 	size_t numval = 0, szval = 0, cursor = 0, slen = 0;
 	cl_device_partition_property *val = NULL;
@@ -1225,7 +1313,7 @@ int device_info_partition_types(cl_device_id dev, cl_device_info param, const ch
 }
 
 int device_info_partition_types_ext(cl_device_id dev, cl_device_info param, const char *pname,
-	const struct device_info_checks *chk UNUSED)
+	const struct device_info_checks* UNUSED(chk))
 {
 	size_t numval = 0, szval = 0, cursor = 0, slen = 0;
 	cl_device_partition_property_ext *val = NULL;
@@ -1283,15 +1371,13 @@ int device_info_partition_types_ext(cl_device_id dev, cl_device_info param, cons
 
 /* Device partition affinity domains */
 int device_info_partition_affinities(cl_device_id dev, cl_device_info param, const char *pname,
-	const struct device_info_checks *chk UNUSED)
+	const struct device_info_checks* UNUSED(chk))
 {
 	cl_device_affinity_domain val;
 	GET_VAL;
 	if (!had_error && val) {
 		/* iterate over affinity domain strings appending their textual form
-		 * to strbuf
-		 * TODO: check for extra bits/no bits
-		 */
+		 * to strbuf */
 		size_t szval = 0;
 		cl_uint i = 0;
 		const char * const *affstr = (output_mode == CLINFO_HUMAN ?
@@ -1307,6 +1393,15 @@ int device_info_partition_affinities(cl_device_id dev, cl_device_info param, con
 			if (szval >= bufsz)
 				break;
 		}
+		/* check for extra bits */
+		if (szval < bufsz) {
+			cl_device_affinity_domain known_mask = ((cl_device_affinity_domain)(1) << affinity_domain_count) - 1;
+			cl_device_affinity_domain extra = val & ~known_mask;
+			if (extra) {
+				add_separator(&szval);
+				szval += snprintf(strbuf + szval, bufsz - szval - 1, "0x%" PRIX64, extra);
+			}
+		}
 	}
 	if (val || had_error)
 		show_strbuf(pname, 0);
@@ -1314,7 +1409,7 @@ int device_info_partition_affinities(cl_device_id dev, cl_device_info param, con
 }
 
 int device_info_partition_affinities_ext(cl_device_id dev, cl_device_info param, const char *pname,
-	const struct device_info_checks *chk UNUSED)
+	const struct device_info_checks* UNUSED(chk))
 {
 	size_t numval = 0, szval = 0, cursor = 0, slen = 0;
 	cl_device_partition_property_ext *val = NULL;
@@ -1489,7 +1584,7 @@ int device_info_qprop(cl_device_id dev, cl_device_info param, const char *pname,
 
 /* Execution capbilities */
 int device_info_execap(cl_device_id dev, cl_device_info param, const char *pname,
-	const struct device_info_checks *chk UNUSED)
+	const struct device_info_checks* UNUSED(chk))
 {
 	cl_device_exec_capabilities val = 0;
 	GET_VAL;
@@ -1516,7 +1611,7 @@ int device_info_execap(cl_device_id dev, cl_device_info param, const char *pname
 
 /* Arch bits and endianness (HUMAN) */
 int device_info_arch(cl_device_id dev, cl_device_info param, const char *pname,
-	const struct device_info_checks *chk UNUSED)
+	const struct device_info_checks* UNUSED(chk))
 {
 	cl_uint bits = 0;
 	{
@@ -1542,8 +1637,8 @@ int device_info_svm_cap(cl_device_id dev, cl_device_info param, const char *pnam
 	const struct device_info_checks *chk)
 {
 	cl_device_svm_capabilities val = 0;
-	int is_20 = dev_is_20(chk);
-	int has_svm_ext = dev_has_svm_ext(chk);
+	const int is_20 = dev_is_20(chk);
+	const int has_amd_svm = (param == CL_DEVICE_SVM_CAPABILITIES && dev_has_amd_svm(chk));
 
 	GET_VAL;
 
@@ -1553,12 +1648,12 @@ int device_info_svm_cap(cl_device_id dev, cl_device_info param, const char *pnam
 		const char * const *scstr = (output_mode == CLINFO_HUMAN ?
 			svm_cap_str : svm_cap_raw_str);
 		set_separator(vbar_str);
-		if (output_mode == CLINFO_HUMAN) {
+		if (output_mode == CLINFO_HUMAN && param == CL_DEVICE_SVM_CAPABILITIES) {
 			/* show 'why' it's being shown */
 			szval += sprintf(strbuf, "(%s%s%s)",
 				(is_20 ? core : empty_str),
-				(is_20 && has_svm_ext ? comma_str : empty_str),
-				chk->has_svm_ext);
+				(is_20 && has_amd_svm ? comma_str : empty_str),
+				chk->has_amd_svm);
 		}
 		for (i = 0; i < svm_cap_count; ++i) {
 			cl_device_svm_capabilities cur = (cl_device_svm_capabilities)(1) << i;
@@ -1576,6 +1671,46 @@ int device_info_svm_cap(cl_device_id dev, cl_device_info param, const char *pnam
 	return had_error;
 }
 
+/* Device terminate capability */
+int device_info_terminate_capability(cl_device_id dev, cl_device_info param, const char *pname,
+	const struct device_info_checks* UNUSED(chk))
+{
+	cl_device_terminate_capability_khr val;
+	GET_VAL;
+	if (!had_error && val) {
+		/* iterate over terminate capability strings appending their textual form
+		 * to strbuf */
+		size_t szval = 0;
+		cl_uint i = 0;
+		const char * const *capstr = (output_mode == CLINFO_HUMAN ?
+			terminate_capability_str : terminate_capability_raw_str);
+		set_separator(output_mode == CLINFO_HUMAN ? comma_str : vbar_str);
+		for (i = 0; i < terminate_capability_count; ++i) {
+			cl_device_terminate_capability_khr cur = (cl_device_terminate_capability_khr)(1) << i;
+			if (val & cur) {
+				/* match: add separator if not first match */
+				add_separator(&szval);
+				szval += bufcpy(szval, capstr[i]);
+			}
+			if (szval >= bufsz)
+				break;
+		}
+		/* check for extra bits */
+		if (szval < bufsz) {
+			cl_device_terminate_capability_khr known_mask = ((cl_device_terminate_capability_khr)(1) << terminate_capability_count) - 1;
+			cl_device_terminate_capability_khr extra = val & ~known_mask;
+			if (extra) {
+				add_separator(&szval);
+				szval += snprintf(strbuf + szval, bufsz - szval - 1, "0x%" PRIX64, extra);
+			}
+		}
+	}
+	if (val || had_error)
+		show_strbuf(pname, 0);
+	return had_error;
+}
+
+
 /*
  * Device info traits
  */
@@ -1606,6 +1741,9 @@ struct device_info_traits dinfo_traits[] = {
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_OPENCL_C_VERSION, "Device OpenCL C Version", str), NULL },
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_EXTENSIONS, "Device Extensions", str_get), NULL },
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_TYPE, "Device Type", devtype), NULL },
+
+	{ CLINFO_BOTH, DINFO(CL_DEVICE_AVAILABLE, "Device Available", bool), NULL },
+
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_PROFILE, "Device Profile", str), NULL },
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_BOARD_NAME_AMD, "Device Board Name (AMD)", str), dev_has_amd },
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_TOPOLOGY_AMD, "Device Topology (AMD)", devtopo_amd), dev_has_amd },
@@ -1647,10 +1785,15 @@ struct device_info_traits dinfo_traits[] = {
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS, "Max work item dimensions", int), NULL },
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_MAX_WORK_ITEM_SIZES, "Max work item sizes", szptr), NULL },
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_MAX_WORK_GROUP_SIZE, "Max work group size", sz), NULL },
-	{ CLINFO_BOTH, DINFO(CL_KERNEL_PREFERRED_WORK_GROUP_SIZE_MULTIPLE, "Preferred work group size multiple", wg), NULL },
+
+	{ CLINFO_BOTH, DINFO(CL_DEVICE_COMPILER_AVAILABLE, "Compiler Available", bool), NULL },
+	{ CLINFO_BOTH, DINFO(CL_DEVICE_LINKER_AVAILABLE, "Linker Available", bool), dev_is_12 },
+	{ CLINFO_BOTH, DINFO(CL_KERNEL_PREFERRED_WORK_GROUP_SIZE_MULTIPLE, "Preferred work group size multiple", wg), dev_has_compiler },
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_WARP_SIZE_NV, "Warp size (NV)", int), dev_has_nv },
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_WAVEFRONT_WIDTH_AMD, "Wavefront width (AMD)", int), dev_is_gpu_amd },
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_MAX_NUM_SUB_GROUPS, "Max sub-groups per work group", int), dev_is_21 },
+	{ CLINFO_BOTH, DINFO(CL_DEVICE_MAX_NAMED_BARRIER_COUNT_KHR, "Max named sub-group barriers", int), dev_has_subgroup_named_barrier },
+	{ CLINFO_BOTH, DINFO(CL_DEVICE_SUB_GROUP_SIZES_INTEL, "Sub-group sizes (Intel)", szptr), dev_has_intel_required_subgroup_size },
 
 	/* Preferred/native vector widths: header is only presented in HUMAN case, that also pairs
 	 * PREFERRED and NATIVE in a single line */
@@ -1693,6 +1836,7 @@ struct device_info_traits dinfo_traits[] = {
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_INTEGRATED_MEMORY_NV, "Integrated memory (NV)", bool), dev_has_nv },
 
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_SVM_CAPABILITIES, "Shared Virtual Memory (SVM) capabilities", svm_cap), dev_has_svm },
+	{ CLINFO_BOTH, DINFO(CL_DEVICE_SVM_CAPABILITIES_ARM, "Shared Virtual Memory (SVM) capabilities (ARM)", svm_cap), dev_has_arm_svm },
 
 	/* Alignment */
 	{ CLINFO_BOTH, DINFO_SFX(CL_DEVICE_MIN_DATA_TYPE_ALIGN_SIZE, "Minimum alignment for any data type", bytes_str, int), NULL },
@@ -1714,8 +1858,8 @@ struct device_info_traits dinfo_traits[] = {
 
 	/* Global memory cache */
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_GLOBAL_MEM_CACHE_TYPE, "Global Memory cache type", cachetype), NULL },
-	{ CLINFO_BOTH, DINFO(CL_DEVICE_GLOBAL_MEM_CACHE_SIZE, "Global Memory cache size", sz), dev_has_cache },
-	{ CLINFO_BOTH, DINFO_SFX(CL_DEVICE_GLOBAL_MEM_CACHELINE_SIZE, "Global Memory cache line", " bytes", int), dev_has_cache },
+	{ CLINFO_BOTH, DINFO(CL_DEVICE_GLOBAL_MEM_CACHE_SIZE, "Global Memory cache size", mem), dev_has_cache },
+	{ CLINFO_BOTH, DINFO_SFX(CL_DEVICE_GLOBAL_MEM_CACHELINE_SIZE, "Global Memory cache line size", " bytes", int), dev_has_cache },
 
 	/* Image support */
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_IMAGE_SUPPORT, "Image support", bool), NULL },
@@ -1729,6 +1873,9 @@ struct device_info_traits dinfo_traits[] = {
 	{ CLINFO_HUMAN, DINFO_SFX(CL_DEVICE_IMAGE2D_MAX_HEIGHT, INDENT "Max 2D image size",  pixels_str, img_sz_2d), dev_has_images },
 	{ CLINFO_RAW, DINFO(CL_DEVICE_IMAGE2D_MAX_HEIGHT, INDENT "Max 2D image height",  sz), dev_has_images },
 	{ CLINFO_RAW, DINFO(CL_DEVICE_IMAGE2D_MAX_WIDTH, INDENT "Max 2D image width",  sz), dev_has_images },
+	{ CLINFO_HUMAN, DINFO_SFX(CL_DEVICE_PLANAR_YUV_MAX_HEIGHT_INTEL, INDENT "Max planar YUV image size",  pixels_str, img_sz_2d), dev_has_intel_planar_yuv },
+	{ CLINFO_RAW, DINFO(CL_DEVICE_PLANAR_YUV_MAX_HEIGHT_INTEL, INDENT "Max planar YUV image height",  sz), dev_has_intel_planar_yuv },
+	{ CLINFO_RAW, DINFO(CL_DEVICE_PLANAR_YUV_MAX_WIDTH_INTEL, INDENT "Max planar YUV image width",  sz), dev_has_intel_planar_yuv },
 	{ CLINFO_HUMAN, DINFO_SFX(CL_DEVICE_IMAGE3D_MAX_HEIGHT, INDENT "Max 3D image size",  pixels_str, img_sz_3d), dev_has_images },
 	{ CLINFO_RAW, DINFO(CL_DEVICE_IMAGE3D_MAX_HEIGHT, INDENT "Max 3D image height",  sz), dev_has_images },
 	{ CLINFO_RAW, DINFO(CL_DEVICE_IMAGE3D_MAX_WIDTH, INDENT "Max 3D image width",  sz), dev_has_images },
@@ -1766,6 +1913,9 @@ struct device_info_traits dinfo_traits[] = {
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_MAX_ON_DEVICE_QUEUES, "Max queues on device", int), dev_is_20 },
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_MAX_ON_DEVICE_EVENTS, "Max events on device", int), dev_is_20 },
 
+	/* Terminate context */
+	{ CLINFO_BOTH, DINFO(CL_DEVICE_TERMINATE_CAPABILITY_KHR, "Terminate capability", terminate_capability), dev_has_terminate_context },
+
 	/* Interop */
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_PREFERRED_INTEROP_USER_SYNC, "Prefer user sync for interop", bool), dev_is_12 },
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_NUM_SIMULTANEOUS_INTEROPS_INTEL, "Number of simulataneous interops (Intel)", int), dev_has_simultaneous_sharing },
@@ -1774,7 +1924,7 @@ struct device_info_traits dinfo_traits[] = {
 	 */
 
 	/* Profiling resolution */
-	{ CLINFO_BOTH, DINFO_SFX(CL_DEVICE_PROFILING_TIMER_RESOLUTION, "Profiling timer resolution", "ns", long), NULL },
+	{ CLINFO_BOTH, DINFO_SFX(CL_DEVICE_PROFILING_TIMER_RESOLUTION, "Profiling timer resolution", "ns", sz), NULL },
 	{ CLINFO_HUMAN, DINFO(CL_DEVICE_PROFILING_TIMER_OFFSET_AMD, "Profiling timer offset since Epoch (AMD)", time_offset), dev_has_amd },
 	{ CLINFO_RAW, DINFO(CL_DEVICE_PROFILING_TIMER_OFFSET_AMD, "Profiling timer offset since Epoch (AMD)", long), dev_has_amd },
 
@@ -1791,13 +1941,12 @@ struct device_info_traits dinfo_traits[] = {
 	 */
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_IL_VERSION, INDENT "IL version", str), dev_is_21, },
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_SPIR_VERSIONS, INDENT "SPIR versions", str), dev_has_spir },
-	{ CLINFO_BOTH, DINFO(CL_DEVICE_PRINTF_BUFFER_SIZE, "printf() buffer size", mem), dev_is_12 },
+	{ CLINFO_BOTH, DINFO(CL_DEVICE_PRINTF_BUFFER_SIZE, "printf() buffer size", mem_sz), dev_is_12 },
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_BUILT_IN_KERNELS, "Built-in kernels", str), dev_is_12 },
-	{ CLINFO_BOTH, DINFO(CL_DEVICE_ME_VERSION_INTEL, "Motion Estimation accelerator version	(Intel)", int), dev_has_intel_AME },
-
-	{ CLINFO_BOTH, DINFO(CL_DEVICE_AVAILABLE, "Device Available", bool), NULL },
-	{ CLINFO_BOTH, DINFO(CL_DEVICE_COMPILER_AVAILABLE, "Compiler Available", bool), NULL },
-	{ CLINFO_BOTH, DINFO(CL_DEVICE_LINKER_AVAILABLE, "Linker Available", bool), dev_is_12 },
+	{ CLINFO_BOTH, DINFO(CL_DEVICE_ME_VERSION_INTEL, "Motion Estimation accelerator version (Intel)", int), dev_has_intel_AME },
+	{ CLINFO_BOTH, DINFO(CL_DEVICE_AVC_ME_VERSION_INTEL, INDENT "Device-side AVC Motion Estimation version", int), dev_has_intel_AVC_ME },
+	{ CLINFO_BOTH, DINFO(CL_DEVICE_AVC_ME_SUPPORTS_TEXTURE_SAMPLER_USE_INTEL, INDENT INDENT "Supports texture sampler use", bool), dev_has_intel_AVC_ME },
+	{ CLINFO_BOTH, DINFO(CL_DEVICE_AVC_ME_SUPPORTS_PREEMPTION_INTEL, INDENT INDENT "Supports preemption", bool), dev_has_intel_AVC_ME },
 };
 
 /* Process all the device info in the traits, except if param_whitelist is not NULL,
@@ -1899,6 +2048,10 @@ printDeviceInfo(const cl_device_id *device, cl_uint d,
 			/* strbuf was abused to give us boolean value */
 			memcpy(&(chk.image_support), strbuf, sizeof(chk.image_support));
 			break;
+		case CL_DEVICE_COMPILER_AVAILABLE:
+			/* strbuf was abused to give us boolean value */
+			memcpy(&(chk.compiler_available), strbuf, sizeof(chk.compiler_available));
+			break;
 		default:
 			/* do nothing */
 			break;
@@ -2222,7 +2375,7 @@ cl_uint checkNullGetDevices(void)
 			}
 		}
 		if (i == num_platforms) {
-			sprintf(strbuf, "<error: platform 0x%p not found>", plat);
+			sprintf(strbuf, "<error: platform 0x%p not found>", (void*)plat);
 		}
 	}
 	printf(I1_STR "%s\n",
@@ -2274,7 +2427,7 @@ void checkNullCtxFromType(void)
 	ALLOC(devs, ndevs, "context devices");
 
 	current_function = __func__;
-	for (t = 2; t < devtype_count; ++t) { /* we skip 0 and _DEFAULT */
+	for (t = 1; t < devtype_count; ++t) { /* we skip 0 */
 		current_param = device_type_raw_str[t];
 
 		sprintf(strbuf, "clCreateContextFromType(NULL, %s)", current_param);
@@ -2333,7 +2486,7 @@ void checkNullCtxFromType(void)
 					break;
 			}
 			if (i == num_platforms) {
-				sprintf(strbuf, "<error: platform 0x%p not found>", plat);
+				sprintf(strbuf, "<error: platform 0x%p not found>", (void*)plat);
 				break;
 			} else {
 				szval += sprintf(strbuf, "%s (%" PRIuS ")",
@@ -2458,22 +2611,14 @@ struct icd_loader_test {
 	{ 12, "clCreateImage" },
 	{ 20, "clSVMAlloc" },
 	{ 21, "clGetHostTimer" },
+	{ 22, "clSetProgramSpecializationConstant" },
 	{ 0, NULL }
 };
 
 int
 icdl_info_str(cl_icdl_info param, const char* pname)
 {
-	error = clGetICDLoaderInfoOCLICD(param, 0, NULL, &nusz);
-	if (nusz > bufsz) {
-		REALLOC(strbuf, nusz, current_param);
-		bufsz = nusz;
-	}
-	had_error = REPORT_ERROR2("get %s size");
-	if (!had_error) {
-		error = clGetICDLoaderInfoOCLICD(param, bufsz, strbuf, NULL);
-		had_error = REPORT_ERROR2("get %s");
-	}
+	GET_STRING2(clGetICDLoaderInfoOCLICD, param);
 	show_strbuf(pname, 1);
 	return had_error;
 }
@@ -2510,7 +2655,7 @@ void oclIcdProps(void)
 		struct icd_loader_test check = icd_loader_tests[i];
 		if (check.symbol == NULL)
 			break;
-		if (dlsym(RTLD_DEFAULT, check.symbol) == NULL)
+		if (dlsym(DL_MODULE, check.symbol) == NULL)
 			break;
 		icdl_ocl_version_found = check.version;
 		++i;
@@ -2528,18 +2673,15 @@ void oclIcdProps(void)
 	 */
 
 #ifdef _MSC_VER
-#pragma warning(push)
-#pragma warning(disable : 4996)
+#pragma warning(suppress : 4996)
 #elif defined __GNUC__ && ((__GNUC__*10 + __GNUC_MINOR__) >= 46)
 #pragma GCC diagnostic push
 #pragma GCC diagnostic ignored "-Wdeprecated-declarations"
 #endif
 
-	clGetICDLoaderInfoOCLICD = clGetExtensionFunctionAddress("clGetICDLoaderInfoOCLICD");
+	PTR_FUNC_PTR clGetICDLoaderInfoOCLICD = clGetExtensionFunctionAddress("clGetICDLoaderInfoOCLICD");
 
-#ifdef _MSC_VER
-#pragma warning(pop)
-#elif defined __GNUC__ && ((__GNUC__*10 + __GNUC_MINOR__) >= 46)
+#if defined __GNUC__ && ((__GNUC__*10 + __GNUC_MINOR__) >= 46)
 #pragma GCC diagnostic pop
 #endif
 
@@ -2598,7 +2740,7 @@ void oclIcdProps(void)
 
 void version(void)
 {
-	puts("clinfo version 2.1.16.01.12");
+	puts("clinfo version 2.2.17.08.25");
 }
 
 void usage(void)
