@@ -4,8 +4,18 @@
  * TODO could be improved by version-checking for C99 support
  */
 
-// also disable strncpy vs strncpy_s warning
+// disable warning about unsafe strncpy vs strncpy_s usage
 #pragma warning(disable : 4996)
+// disable warning about constant conditional expressions
+#pragma warning(disable : 4127)
+// disable warning about non-constant aggregate initializer
+#pragma warning(disable:  4204)
+
+// Suppress warning about unused parameters. The macro definition
+// _should_ work, but it doesn't on VS2012 (cl 17), may be a version thing
+#define UNUSED(x) x __pragma(warning(suppress: 4100))
+// TODO FIXME remove full-blown warning removal where not needed
+#pragma warning(disable: 4100)
 
 // No inline in MS C
 #define inline __inline
